
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#pragma once

            #include <glib-object.h>
            #include <gst/transcoder/transcoder-prelude.h>

            G_BEGIN_DECLS

/* enumerations from "gsttranscoder.h" */

GST_TRANSCODER_API
GType gst_transcoder_error_get_type (void);
#define GST_TYPE_TRANSCODER_ERROR (gst_transcoder_error_get_type())

GST_TRANSCODER_API
GType gst_transcoder_state_get_type (void);
#define GST_TYPE_TRANSCODER_STATE (gst_transcoder_state_get_type())

GST_TRANSCODER_API
GType gst_transcoder_message_get_type (void);
#define GST_TYPE_TRANSCODER_MESSAGE (gst_transcoder_message_get_type())

G_END_DECLS

/* Generated data ends here */

